//@line 36 "/builds/tinderbox/Sb-Mozilla1.8-l10n/Linux_2.4.21-37.EL_Clobber/mozilla/calendar/locales/../../../l10n/mk/calendar/sunbird-l10n.js"

//@line 38 "/builds/tinderbox/Sb-Mozilla1.8-l10n/Linux_2.4.21-37.EL_Clobber/mozilla/calendar/locales/../../../l10n/mk/calendar/sunbird-l10n.js"

//@line 41 "/builds/tinderbox/Sb-Mozilla1.8-l10n/Linux_2.4.21-37.EL_Clobber/mozilla/calendar/locales/../../../l10n/mk/calendar/sunbird-l10n.js"
pref("calendar.week.start", 0);

//@line 44 "/builds/tinderbox/Sb-Mozilla1.8-l10n/Linux_2.4.21-37.EL_Clobber/mozilla/calendar/locales/../../../l10n/mk/calendar/sunbird-l10n.js"
pref("calendar.week.d0sundaysoff", true);
pref("calendar.week.d1mondaysoff", false);
pref("calendar.week.d2tuesdaysoff", false);
pref("calendar.week.d3wednesdaysoff", false);
pref("calendar.week.d4thursdaysoff", false);
pref("calendar.week.d5fridaysoff", false);
pref("calendar.week.d6saturdaysoff", true);

pref("general.useragent.locale", "mk");

//@line 55 "/builds/tinderbox/Sb-Mozilla1.8-l10n/Linux_2.4.21-37.EL_Clobber/mozilla/calendar/locales/../../../l10n/mk/calendar/sunbird-l10n.js"
pref("calendar.categories.names", "Годишнина,Роденден,Службено,Повици,Клиенти,Конкуренција,Муштерија,Омилени,За проверка,Поклони,Празници,Идеи,Проблеми,Разно,Приватно,Проекти,Јавен празник,Статус,Добавувачи,Патување,Одмор");
